from typing import Callable, Iterable, Optional, Protocol, Union

import numpy as np
from scipy.spatial.distance import pdist

from experiment_design.variable import DesignSpace
from experiment_design.variable.space import VariableCollection


class Scorer(Protocol):
    def __call__(self, doe: np.ndarray) -> float:
        """
        A scoring function to evaluate an experiment design quality. Larger values are better,
        i.e. this will be maximized.

        :param doe: Design of experiments consisting of candidate samples.
        :return: score of the DoE.
        """
        ...


class ScorerFactory(Protocol):
    def __call__(
        self,
        variables: VariableCollection,
        sample_size: int,
        old_sample: Optional[np.ndarray] = None,
    ) -> Scorer:
        """Given variables and sample size, create a scoring function"""
        ...


class MaxCorrelationScorerFactory:
    """
    A scorer factory for the maximum absolute correlation error between sampling points.

    :param target_correlation: A symmetric matrix with shape (len(variables), len(variables)), representing the linear
        dependency between the dimensions.
    :param local: If True, any points in the old_sample will be ignored, that fall outside the finite bounds of the
        provided variables. Has no effect if old_sample is None.
    :param eps: a small positive value to improve the stability of the log operation.

    """

    def __init__(
        self,
        target_correlation: Union[float, np.ndarray] = 0,
        local: bool = True,
        eps: float = 1e-2,
    ) -> None:
        if np.max(np.abs(target_correlation)) > 1:
            raise ValueError("Correlations should be in the interval [-1,1].")
        self.target_correlation = target_correlation

        self.local = local

        if eps < 0:
            raise ValueError("eps must be positive")
        self.eps = eps

    def __call__(
        self,
        variables: VariableCollection,
        sample_size: int,
        old_sample: Optional[np.ndarray] = None,
    ) -> Scorer:
        """
        Creates a scorer, that computes the maximum absolute correlation error between the candidate samples
        and the target_correlation.

        :param variables: Dimensions of the design space.
        :param sample_size: number of candidate points to be scored.
        :param old_sample: If passed, represents the matrix of points in an older design of experiments with shape
            (old_sample_size, len(variables)). Depending on self.local, some or all of these will be appended to the
            candidate points before computing the correlation error.
        :return: a scorer that returns the negative exp(maximum absolute correlation error + 1). If the error is
            smaller than self.eps, it returns negative exp(maximum absolute correlation error - 1) instead.
        """

        target_correlation = create_correlation_matrix(
            self.target_correlation, num_variables=len(variables)
        )
        handler = create_old_doe_handler(variables, old_sample, local=self.local)

        def _scorer(doe: np.ndarray) -> float:
            error = np.max(
                np.abs(np.corrcoef(handler(doe), rowvar=False) - target_correlation)
            )
            if error > self.eps:
                return -np.exp(error + 1)
            return -np.exp(error - 1)

        return _scorer


class PairwiseDistanceScorerFactory:
    """
    A scorer factory for the minimum pairwise distance between sampling points.

    Note: This currently computes all distances. Hence, it may be inefficient for very large
    sample sizes. This behaviour could be improved using algorithms like KDTrees or KNN but
    they would lead to more dependencies. Therefore, we prefer omitting the implementation for
    now and providing this note instead. If there is interest in handling large sample sizes,
    create an issue or open a pull request.

    :param local: If True, any points in the old_sample will be ignored, that fall outside the finite bounds of the
        provided variables. Has no effect if old_sample is None.
    :return: a scorer that returns the log minimum pairwise distance divided by the log max distance.
    """

    def __init__(self, local: bool = False) -> None:
        self.local = local

    def __call__(
        self,
        variables: VariableCollection,
        sample_size: int,
        old_sample: Optional[np.ndarray] = None,
    ) -> Scorer:
        """
        Create a scorer, that computes the minimum pairwise distance between sampling points.s

        :param variables: Dimensions of the design space.
        :param sample_size: number of candidate points to be scored.
        :param old_sample: If passed, represents the matrix of points in an older design of experiments with shape
            (old_sample_size, len(variables)). Depending on self.local, some or all of these will be appended to the
            candidate points before computing the correlation error.
        :return: a scorer that returns the log minimum pairwise distance divided by the log max distance,
        """

        handler = create_old_doe_handler(variables, old_sample, local=self.local)
        bin_diagonal_length = calculate_equidistant_bin_diagonal_length(
            variables, sample_size
        )

        def _scorer(doe: np.ndarray) -> float:
            min_pairwise_distance = np.min(pdist(handler(doe)))
            return np.exp(min_pairwise_distance / bin_diagonal_length)

        return _scorer


class WeightedSumScorerFactory:
    """
    A factory that creates a weighted sum of multiple scorers

    :param scorer_factories: These are combined by adding the scores their scorers provide.
    :param weights: Weights to use for combining the scorers. If not passed, each scores will not be weighed.
    """

    def __init__(
        self, scorer_factories: list[ScorerFactory], weights: Iterable[float]
    ) -> None:
        self.scorer_factories = scorer_factories
        if weights is None:
            weights = np.ones(len(scorer_factories))
        self.weights = np.array(weights).ravel()
        if self.weights.size != len(self.scorer_factories):
            raise ValueError(
                f"Wrong number of scorer_factories ({len(scorer_factories)}) and weights ({self.weights.size})"
            )

    def __call__(
        self,
        variables: VariableCollection,
        sample_size: int,
        old_sample: Optional[np.ndarray] = None,
    ) -> Scorer:

        scorers = [
            factory(variables, sample_size, old_sample=old_sample)
            for factory in self.scorer_factories
        ]

        def _scorer(doe: np.ndarray) -> float:
            return sum(
                [scorer(doe) * weight for scorer, weight in zip(scorers, self.weights)]
            )

        return _scorer


def create_default_scorer_factory(
    target_correlation: Union[np.ndarray, float] = 0.0,
    distance_score_weight: float = 0.9,
    correlation_score_weight: float = 0.1,
    local_correlation: bool = True,
    local_pairwise_distance: bool = False,
) -> ScorerFactory:
    """
    Create a scorer factory, which is a weighted sum of maximum correlation error and
    minimum pairwise distance scorers

    :param target_correlation: A symmetric matrix with shape (len(variables), len(variables)), representing the linear
        dependency between the dimensions. If a float, all non-diagonal entries of the unit matrix will be set to this
        value.
    :param distance_score_weight: Weight of the minimum pairwise distance score.
    :param correlation_score_weight: Weight of the maximum correlation error score.
    :param local_correlation: Controls the local attribute of the MaxCorrelationScorerFactory.
    :param local_pairwise_distance: Controls the local attribute of the PairwiseDistanceScorerFactory
    :return: WeightedSumScorerFactory instance.
    """
    corr_scorer_factory = MaxCorrelationScorerFactory(
        target_correlation=target_correlation, local=local_correlation
    )
    dist_scorer_factory = PairwiseDistanceScorerFactory(local=local_pairwise_distance)
    return WeightedSumScorerFactory(
        scorer_factories=[corr_scorer_factory, dist_scorer_factory],
        weights=[correlation_score_weight, distance_score_weight],
    )


def calculate_equidistant_bin_diagonal_length(
    variables: VariableCollection, sample_size: int
) -> float:
    """Calculate the length of the diagonal of equidistant bins the (Euclidean) design space"""
    if not isinstance(variables, DesignSpace):
        variables = DesignSpace(variables)
    lower, upper = variables.lower_bound, variables.upper_bound
    return np.linalg.norm((np.array(upper) - np.array(lower)) / sample_size)


def create_correlation_matrix(
    target_correlation: Union[np.ndarray, float] = 0.0,
    num_variables: Optional[int] = None,
) -> np.ndarray:
    """Create a correlation matrix from the target correlation in case it is a float"""
    if not np.isscalar(target_correlation):
        return target_correlation
    if not num_variables:
        raise ValueError(
            "num_variables have to be passed if the target_correlation is a scalar."
        )
    return (
        np.eye(num_variables) * (1 - target_correlation)
        + np.ones((num_variables, num_variables)) * target_correlation
    )


def create_old_doe_handler(
    variables: VariableCollection,
    old_sample: Optional[np.ndarray] = None,
    local: bool = False,
) -> Callable[[np.ndarray], np.ndarray]:
    """
    Return a function to handle any old design of experiments. Specifically,
    append some or all of the points from an old design of experiments (DoE) for
    including them in the scoring.

    :param variables: Dimensions of the design space. Only relevant if local is set to True.
    :param old_sample: Matrix of points with shape=(sample_size, len(variables)) in the old DoE. If None, this returns
        a no-op function.
    :param local: If True, only include the points from the old_doe, that fall between the finite local bounds of the
        variables.
    :return: The function with a single argument, new doe, which may append points from the old doe depending on the
        arguments provided.
    """
    if old_sample is None:
        # Nothing to handle
        return lambda x: x

    if not local:
        # Append every point in the old doe
        return lambda x: np.append(old_sample, x, axis=0)

    old_sample = select_local(old_sample, variables)
    return lambda x: np.append(old_sample, x, axis=0)


def select_local(samples: np.ndarray, variables: VariableCollection) -> np.ndarray:
    """Select and return samples that fall within the finite bounds of the variables"""
    if not isinstance(variables, DesignSpace):
        variables = DesignSpace(variables)
    lower, upper = variables.lower_bound[None, :], variables.upper_bound[None, :]
    local_mask = np.logical_and((samples >= lower).all(1), (samples <= upper).all(1))
    return samples[local_mask]



from functools import partial
from typing import Callable, Optional, Union

import numpy as np
from scipy.linalg import solve_triangular
from scipy.stats import uniform

from experiment_design.experiment_designer import ExperimentDesigner
from experiment_design.optimize import (
    random_search,
    simulated_annealing_by_perturbation,
)
from experiment_design.scorers import (
    Scorer,
    ScorerFactory,
    create_correlation_matrix,
    create_default_scorer_factory,
    select_local,
)
from experiment_design.variable import DesignSpace, VariableCollection


class OrthogonalSamplingDesigner(ExperimentDesigner):
    """
    Create or extend an orthogonal sampling design. Orthogonal sampling design partitions the design space into
    bins of equal marginal probability and places samples such that each bin is only filled once for each dimension.
    If all variables are uniform, orthogonal sampling becomes a Latin hypercube sampling.

    :param target_correlation: A symmetric matrix with shape (len(variables), len(variables)), representing the linear
        dependency between the dimensions. If a float, all non-diagonal entries of the unit matrix will be set to this
        value.
    :param inter_bin_randomness: Controls the randomness of placed points between the bin bounds. Specifically, 0. means that
        the points are placed at the center of each bin, whereas 1. leads to a random point placement within the bounds.
        Any other fractions leads to a random placement within that fraction of the bin bounds in each dimension.

    :param non_occupied_bins: Only relevant for extending the design, i.e. if old points are provided, and if the constraint
        regarding the number of occupation of each bin has to be violated. True means that each bin is occupied at least
        once for each dimension, although some bins might be occupied more often. Otherwise, each bin is occupied once
        or less often, leading to empty bins in some cases.
    :param scorer_factory: A factory that creates scorers for the given variables, sample_size and in the cast of an
        extension, old sampling points. If not passed, a default one will be created, that evaluates the maximum
        correlation error and minimum pairwise distance. See `experiment_design.scorers.create_default_scorer_factory`
        for more details.
    """

    def __init__(
        self,
        target_correlation: Union[np.ndarray, float] = 0.0,
        inter_bin_randomness: float = 1.0,
        non_occupied_bins: bool = False,
        scorer_factory: Optional[ScorerFactory] = None,
    ) -> None:
        self.target_correlation = target_correlation
        self.inter_bin_randomness = inter_bin_randomness
        if non_occupied_bins:
            self.empty_size_check = np.min
        else:
            self.empty_size_check = np.max
        if scorer_factory is None:
            scorer_factory = create_default_scorer_factory(
                target_correlation=target_correlation
            )
        super(OrthogonalSamplingDesigner, self).__init__(scorer_factory=scorer_factory)

    def _create(
        self,
        variables: DesignSpace,
        sample_size: int,
        scorer: Scorer,
        initial_steps: int,
        final_steps: int,
        verbose: int,
    ) -> np.ndarray:
        target_correlation = create_correlation_matrix(
            self.target_correlation, num_variables=variables.dimensions
        )
        if (initial_steps + final_steps) <= 2:
            # Enable faster use cases:
            return create_orthogonal_design(
                variables=variables,
                sample_size=sample_size,
                target_correlation=target_correlation,
                inter_bin_randomness=self.inter_bin_randomness,
            )

        if verbose:
            print("Creating an initial design")
        doe = random_search(
            creator=partial(
                create_orthogonal_design,
                variables=variables,
                sample_size=sample_size,
                target_correlation=target_correlation,
                inter_bin_randomness=self.inter_bin_randomness,
            ),
            scorer=scorer,
            steps=initial_steps,
            verbose=verbose,
        )
        if verbose:
            print("Optimizing the initial design")
        return simulated_annealing_by_perturbation(
            doe, scorer, steps=final_steps, verbose=verbose
        )

    def _extend(
        self,
        old_sample: np.ndarray,
        variables: DesignSpace,
        sample_size: int,
        scorer: Scorer,
        initial_steps: int,
        final_steps: int,
        verbose: int,
    ) -> np.ndarray:
        local_doe = select_local(old_sample, variables)
        probabilities = variables.cdf_of(local_doe)
        if not np.all(np.isfinite(probabilities)):
            raise RuntimeError(
                "Non-finite probability encountered. Please check the distributions."
            )

        bins_per_dimension = sample_size + local_doe.shape[0]

        empty = _find_sufficient_empty_bins(
            probabilities,
            bins_per_dimension,
            sample_size,
            empty_size_check=self.empty_size_check,
        )
        if verbose:
            print("Creating candidate points to extend the design")
        new_sample = random_search(
            creator=partial(
                _create_candidates_from,
                empty_bins=empty,
                variables=variables,
                sample_size=sample_size,
                inter_bin_randomness=self.inter_bin_randomness,
            ),
            scorer=scorer,
            steps=initial_steps,
            verbose=verbose,
        )
        if verbose:
            print("Optimizing candidate points to extend the design")
        return simulated_annealing_by_perturbation(
            new_sample, scorer, steps=final_steps
        )


def create_orthogonal_design(
    variables: VariableCollection,
    sample_size: int,
    target_correlation: np.ndarray,
    inter_bin_randomness: float = 1.0,
) -> np.ndarray:
    """Create an orthogonal design without any optimization."""
    if not isinstance(variables, DesignSpace):
        variables = DesignSpace(variables)
    # Sometimes, we may randomly generate probabilities with
    # singular correlation matrices. Try 3 times to avoid issue until we give up
    for k in range(3):
        probabilities = create_lhd_probabilities(
            len(variables), sample_size, inter_bin_randomness=inter_bin_randomness
        )
        doe = variables.value_of(probabilities)
        try:
            return iman_connover_transformation(doe, target_correlation)
        except np.linalg.LinAlgError:
            pass

    return doe


def create_lhd_probabilities(
    num_variables: int,
    sample_size: int,
    inter_bin_randomness: float = 1.0,
) -> np.ndarray:
    """Create probabilities for a Latin hypercube design."""
    if not 0.0 <= inter_bin_randomness <= 1.0:
        raise ValueError(
            f"inter_bin_randomness has to be between 0 and 1, got {inter_bin_randomness}"
        )
    doe = uniform.rvs(size=(sample_size, num_variables))
    doe = (np.argsort(doe, axis=0) + 0.5) / sample_size
    if inter_bin_randomness == 0.0:
        return doe
    delta = inter_bin_randomness / sample_size
    return doe + uniform(-delta / 2, delta).rvs(size=(sample_size, num_variables))


def iman_connover_transformation(
    doe: np.ndarray,
    target_correlation: np.ndarray,
    means: Optional[np.ndarray] = None,
    standard_deviations: Optional[np.ndarray] = None,
) -> np.ndarray:
    """Rearrange the values of doe to reduce correlation error while keeping the Latin hypercube constraint"""
    # See Chapter 4.3.2 of
    # Local Latin Hypercube Refinement for Uncertainty Quantification and Optimization, Can Bogoclu, (2022)
    if means is None:
        means = np.mean(doe, axis=0)
    if standard_deviations is None:
        standard_deviations = np.std(doe, axis=0, keepdims=True)
        standard_deviations = standard_deviations.reshape((1, -1))
    target_covariance = (
        standard_deviations.T.dot(standard_deviations) * target_correlation
    )

    transformed = second_moment_transformation(doe, means, target_covariance)
    order = np.argsort(np.argsort(transformed, axis=0), axis=0)
    return np.take_along_axis(np.sort(doe, axis=0), order, axis=0)


def second_moment_transformation(
    doe: np.ndarray,
    means: Union[float, np.ndarray],
    target_covariance: np.ndarray,
) -> np.ndarray:
    """Second-moment transformation for achieving the target covariance"""
    target_cov_upper = np.linalg.cholesky(
        target_covariance
    ).T  # convert to covariance before Cholesky
    cur_cov_upper = np.linalg.cholesky(np.cov(doe, rowvar=False)).T
    inv_cov_upper = solve_triangular(cur_cov_upper, target_cov_upper)
    return (doe - means).dot(inv_cov_upper) + means


def _find_sufficient_empty_bins(
    probabilities: np.ndarray,
    bins_per_dimension: int,
    required_sample_size: int,
    empty_size_check: Callable[[np.ndarray], float] = np.max,
) -> np.ndarray:
    empty = _find_empty_bins(probabilities, bins_per_dimension)
    cols = np.where(empty)[1]
    while (
        empty_size_check(np.unique(cols, return_counts=True)[1]) < required_sample_size
    ):
        bins_per_dimension += 1
        empty = _find_empty_bins(probabilities, bins_per_dimension)
        cols = np.where(empty)[1]
    return empty


def _find_empty_bins(probabilities: np.ndarray, bins_per_dimension: int) -> np.ndarray:
    """
    Find empty bins on an orthogonal sampling grid given the probabilities.

    :param probabilities: Array of cdf values of the observed points.
    :param bins_per_dimension: Determines the size of the grid to be tested.
    :return: Boolean array of empty bins with shape=(n_bins, n_dims).
    """
    empty_bins = np.ones((bins_per_dimension, probabilities.shape[1]), dtype=bool)
    edges = np.arange(bins_per_dimension + 1) / bins_per_dimension
    edges = edges.reshape((-1, 1))
    for i_dim in range(probabilities.shape[1]):
        condition = np.logical_and(
            probabilities[:, i_dim] >= edges[:-1], probabilities[:, i_dim] < edges[1:]
        )
        empty_bins[:, i_dim] = np.logical_not(condition.any(1))
    return empty_bins


def _create_candidates_from(
    empty_bins: np.ndarray,
    variables: DesignSpace,
    sample_size: int,
    inter_bin_randomness: float = 1.0,
) -> np.ndarray:
    if not 0.0 <= inter_bin_randomness <= 1.0:
        raise ValueError(
            f"inter_bin_randomness has to be between 0 and 1, got {inter_bin_randomness}"
        )
    empty_rows, empty_cols = np.where(empty_bins)
    bins_per_dimension, dimensions = empty_bins.shape
    delta = 1 / bins_per_dimension
    probabilities = np.empty((sample_size, dimensions))
    for i_dim in range(dimensions):
        values = empty_rows[empty_cols == i_dim]
        np.random.shuffle(values)
        diff = sample_size - values.size
        if diff < 0:
            values = values[:sample_size]
        elif diff > 0:
            available = [idx for idx in range(bins_per_dimension) if idx not in values]
            extra = np.random.choice(available, diff, replace=False)
            values = np.append(extra, values)
        probabilities[:, i_dim] = values * delta + delta / 2
    if inter_bin_randomness > 0.0:
        delta *= inter_bin_randomness
        probabilities += uniform(-delta / 2, delta).rvs(size=(sample_size, dimensions))
    return variables.value_of(probabilities)



import abc
from typing import Optional

import numpy as np

from experiment_design.scorers import (
    Scorer,
    ScorerFactory,
    create_default_scorer_factory,
)
from experiment_design.variable import DesignSpace, VariableCollection

DEFAULT_INITIAL_OPTIMIZATION_PROPORTION = 0.1


class ExperimentDesigner(abc.ABC):
    """
    Abstract class of experiment designers that create and extend designs of experiments, i.e. experiment plans

    :param scorer_factory: A factory that creates scorers for the given variables, sample_size and in the cast of an
        extension, old sampling points. If not passed, a default one will be created, that evaluates the maximum
        correlation error and minimum pairwise distance. See `experiment_design.scorers.create_default_scorer_factory`
        for more details.
    """

    def __init__(self, scorer_factory: Optional[ScorerFactory] = None) -> None:
        if scorer_factory is None:
            scorer_factory = create_default_scorer_factory()
        self.scorer_factory = scorer_factory

    def design(
        self,
        variables: VariableCollection,
        sample_size: int,
        old_sample: Optional[np.ndarray] = None,
        steps: Optional[int] = None,
        initial_optimization_proportion: float = DEFAULT_INITIAL_OPTIMIZATION_PROPORTION,
        verbose: int = 0,
    ) -> np.ndarray:
        """
        Create or extend a design of experiments (DoE).

        :param variables: Determines the dimensions of the resulting sample.
        :param sample_size: the number of points to be created.
        :param old_sample: Old DoE matrix with shape (len(variables), old_sample_size). If provided,
            it will be extended with sample_size new points, otherwise a new DoE will be created.
            In both cases, only the new points will be returned.
        :param steps: Number of DoEs to be created to choose the best from.
        :param initial_optimization_proportion:  Proportion of steps that will be used to create an
            initial DoE with a good score. Rest of the steps will be used to optimize the candidate points.
        :param verbose: Controls print messages. 2 leads to maximum verbosity,
        :return: DoE matrix with shape (len(variables), samples_size)
        """
        if not isinstance(variables, DesignSpace):
            variables = DesignSpace(variables)
        scorer = self.scorer_factory(variables, sample_size, old_sample=old_sample)
        initial_steps, final_steps = calculate_optimization_step_numbers(
            sample_size, steps, proportion=initial_optimization_proportion
        )
        if old_sample is None:
            return self._create(
                variables, sample_size, scorer, initial_steps, final_steps, verbose
            )
        return self._extend(
            old_sample,
            variables,
            sample_size,
            scorer,
            initial_steps,
            final_steps,
            verbose,
        )

    @abc.abstractmethod
    def _create(
        self,
        variables: DesignSpace,
        sample_size: int,
        scorer: Scorer,
        initial_steps: int,
        final_steps: int,
        verbose: int,
    ) -> np.ndarray:
        raise NotImplementedError

    @abc.abstractmethod
    def _extend(
        self,
        old_sample: np.ndarray,
        variables: DesignSpace,
        sample_size: int,
        scorer: Scorer,
        initial_steps: int,
        final_steps: int,
        verbose: int,
    ) -> np.ndarray:
        raise NotImplementedError


def calculate_optimization_step_numbers(
    sample_size: int,
    steps: Optional[int] = None,
    proportion: float = DEFAULT_INITIAL_OPTIMIZATION_PROPORTION,
) -> tuple[int, int]:
    """
    Calculate the proportion of steps to be used for the initial and final optimizations.
    Also provide a sensible default total number of steps if they are not provided

    :param sample_size: The number of samples to be created. Only used to guess steps if it is
        not provided.
    :param steps: The number of total steps. If not passed, it will be guessed depending on the
        sample_size.
    :param proportion: Proportion of initial to total steps.
    :return: initial and final optimization steps (minimum 1 for each stage).
    """
    if steps is None:
        if sample_size <= 128:
            steps = 20000
        else:
            steps = 2000
    init_steps = max(1, round(proportion * steps))
    opt_steps = max(1, steps - init_steps)
    return init_steps, opt_steps



import warnings
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Callable

import numpy as np
from scipy.special import comb as combine

from experiment_design.scorers import Scorer


def random_search(
    creator: Callable[[], np.ndarray], scorer: Scorer, steps: int, verbose: int = 0
) -> np.ndarray:
    """
    Given a design of experiment (DoE) creator and scorer, maximize the score
    by random search.

    :param creator: A function that creates DoEs.
    :param scorer: A function that scores DoEs.
    :param steps: Number of steps to search.
    :param verbose: Controls print messages. 2 leads to maximum verbosity,
    :return: The DoE matrix with the best score.
    """
    steps = max(1, steps)
    best_doe = creator()
    start_score = best_score = scorer(best_doe)
    if verbose > 1:
        print(f"Initial score: {start_score:.2e}")
    for i_try in range(2, steps + 1):
        doe = creator()
        score = scorer(doe)
        if score > best_score:
            best_doe = doe
            best_score = score
            if verbose > 1:
                print(
                    f"Step {i_try} - start score improved by {100 * abs((best_score - start_score) / start_score):.1f}%"
                )
    if verbose:
        print(
            f"Final score improved start score by {100 * abs((best_score - start_score) / start_score):.1f}% in {steps} steps"
        )
    return best_doe


def simulated_annealing_by_perturbation(
    doe: np.ndarray,
    scorer: Scorer,
    steps: int = 1000,
    cooling_rate: float = 0.95,
    temperature: float = 25.0,
    max_steps_without_improvement: int = 25,
    verbose: int = 0,
) -> np.ndarray:
    """
    Simulated annealing algorithm to maximize the score of a design of experiments (DoE) by
    perturbing the rows of the design matrix along the columns. This kind of perturbation is
    used to avoid violating the Latin hypercube scheme, i.e. to keep the number of filled bins
    same.

    :param doe: DoE matrix with shape (sample_size, len(variables)).
    :param scorer: A scoring function for the doe. It will be maximized.
    :param steps: Number of steps for the annealing algorithm.
    :param cooling_rate: Annealing parameter to decay temperature.
    :param temperature: Annealing temperature.
    :param max_steps_without_improvement: A limit on the maximum steps to take for exploration before setting the
        reference matrix to the last best value.
    :param verbose: Controls print messages. 2 leads to maximum verbosity,
    :return: Optimized DoE matrix
    """
    if temperature <= 1e-16:
        raise ValueError("temperature must be strictly positive.")
    if not 0 <= cooling_rate <= 1:
        raise ValueError("decay has to be between 0 and 1.")

    def cool_down_temperature(temp: float, min_temperature: float = 1e-6) -> float:
        return max(temp * cooling_rate, min_temperature)

    if max_steps_without_improvement < 1:
        max_steps_without_improvement = 1

    doe_start = doe.copy()
    best_doe = doe.copy()
    start_score = anneal_step_score = best_score = scorer(doe)
    if verbose > 1:
        print(f"Initial score: {start_score:.2e}")
    steps_without_improvement = 0
    switch_cache = _MatrixRowSwitchCache(
        row_size=doe.shape[0], column_size=doe.shape[1]
    )
    old_switch_cache = _MatrixRowSwitchCache(
        row_size=doe.shape[0], column_size=doe.shape[1]
    )
    for i_try in range(2, steps + 1):
        if switch_cache.is_full:
            if verbose > 1:
                print("No more perturbation left to improve the score")
            break
        doe_try = switch_cache.switch_rows_and_cache(doe_start.copy())
        curr_score = scorer(doe_try)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="overflow")
            transition_probability = np.exp(
                -(anneal_step_score - curr_score) / temperature
            )

        if (
            curr_score > anneal_step_score
            or np.random.random() <= transition_probability
        ):
            doe_start = doe_try.copy()
            old_switch_cache = deepcopy(switch_cache)
            anneal_step_score = curr_score
            switch_cache.reset()
            temperature = cool_down_temperature(temperature)
            if curr_score > best_score:
                best_doe = doe_try.copy()
                best_score = curr_score
                steps_without_improvement = 0
                if verbose > 1:
                    print(
                        f"{i_try} - start score improved by {100 * abs((best_score - start_score) / start_score):.1f}%"
                    )
        steps_without_improvement += 1

        if steps_without_improvement >= max_steps_without_improvement:
            temperature = cool_down_temperature(temperature)
            # Bound Randomness by setting back to best result
            # This often accelerates convergence speed
            doe_start = best_doe.copy()
            switch_cache = deepcopy(old_switch_cache)
            anneal_step_score = best_score
            steps_without_improvement = 0

    if verbose:
        print(
            f"Final score improved start score by {100 * abs((best_score - start_score) / start_score):.1f}% in {steps} steps"
        )
    return best_doe


@dataclass
class _MatrixRowSwitchCache:
    """A cache that is used in simulated annealing to avoid repeating the same perturbations"""

    row_size: int
    column_size: int
    _max_switches_per_column: int = field(init=False, repr=False, default=None)
    _cache: list[np.ndarray] = field(init=False, repr=False, default=None)
    _cache_sizes: np.ndarray = field(init=False, repr=False, default=None)

    def __post_init__(self) -> None:
        # We switch two rows, thus the magic number
        self._max_switches_per_column = combine(self.row_size, 2)
        self.reset()

    @property
    def is_full(self) -> bool:
        return np.min(self._cache_sizes) >= self._max_switches_per_column

    def reset(self) -> None:
        self._cache_sizes = np.zeros(self.column_size, dtype=int)
        self._cache = [np.zeros((0, 2), dtype=int) for _ in range(self.column_size)]

    def choose_column(self) -> int:
        try:
            return np.random.choice(
                np.where(self._cache_sizes < self._max_switches_per_column)[0]
            )
        except:
            print(np.where(self._cache_sizes < self._max_switches_per_column)[0])
            pass

    def choose_rows(self, column: int) -> tuple[int, int]:
        def choose_non_occupied(occupied: set[int]) -> int:
            return np.random.choice(
                [idx for idx in range(self.row_size) if idx not in occupied]
            )

        row_cache = self._cache[column]
        max_switches_per_row = self.row_size - 1
        fulls, counts = np.unique(row_cache, return_counts=True)
        row_1 = choose_non_occupied(set(fulls[counts >= max_switches_per_row]))
        row_2 = choose_non_occupied(
            set(row_cache[np.any(row_cache == row_1, axis=1)].ravel())
        )
        return row_1, row_2

    def cache(self, column: int, row_1: int, row_2: int) -> None:
        self._cache[column] = np.append(
            self._cache[column], np.array([[row_1, row_2]]), axis=0
        )
        self._cache_sizes[column] += 1

    def switch_rows_and_cache(self, matrix: np.ndarray) -> np.ndarray:
        column = self.choose_column()
        row_1, row_2 = self.choose_rows(column)
        self.cache(column, row_1, row_2)
        matrix[row_1, column], matrix[row_2, column] = (
            matrix[row_2, column],
            matrix[row_1, column],
        )
        return matrix



from dataclasses import dataclass, field
from typing import Union

import numpy as np

from experiment_design.variable import Variable


@dataclass
class DesignSpace:
    """A container of multiple variables defining a design space."""

    variables: list[Variable]
    _lower_bound: np.ndarray = field(init=False, repr=False, default=None)
    _upper_bound: np.ndarray = field(init=False, repr=False, default=None)

    def __post_init__(self) -> None:
        lower, upper = [], []
        for var in self.variables:
            lower.append(var.finite_lower_bound)
            upper.append(var.finite_upper_bound)
        self._lower_bound = np.array(lower)
        self._upper_bound = np.array(upper)

    def _map_by(self, attribute: str, values: np.ndarray) -> np.ndarray:
        if len(values.shape) != 2:
            values = values.reshape((-1, len(self.variables)))
        results = np.zeros(values.shape)
        for i_dim, variable in enumerate(self.variables):
            results[:, i_dim] = getattr(variable, attribute)(values[:, i_dim])
        return results

    def value_of(self, probabilities: np.ndarray) -> np.ndarray:
        """Given an array of probabilities return the corresponding values using the inverse cdf."""
        return self._map_by("value_of", probabilities)

    def cdf_of(self, values: np.ndarray) -> np.ndarray:
        """Given an array of values return the probability using the cdf."""
        return self._map_by("cdf_of", values)

    @property
    def lower_bound(self) -> np.ndarray:
        """Finite lower bound of the space."""
        return self._lower_bound

    @property
    def upper_bound(self) -> np.ndarray:
        """Finite upper bound of the space."""
        return self._upper_bound

    @property
    def dimensions(self) -> int:
        """Size of the space, i.e. the number of dimensions."""
        return len(self.variables)

    def __len__(self):
        """Size of the space, i.e. the number of dimensions."""
        return self.dimensions


VariableCollection = Union[list[Variable], DesignSpace]



from dataclasses import dataclass
from typing import Any, Callable, Optional, Protocol, Sequence, Union

import numpy as np
from scipy.stats import randint, rv_continuous, rv_discrete, uniform

# Following is ugly, but it is scipy's fault for not exposing rv_frozen
# noinspection PyProtectedMember
from scipy.stats._distn_infrastructure import rv_frozen


def _is_frozen_discrete(dist: Any) -> bool:
    """Check if dist is a rv_frozen_discrete instance"""
    return isinstance(dist, rv_frozen) and isinstance(dist.dist, rv_discrete)


def _is_frozen_continuous(dist: Any) -> bool:
    """Check if dist is a rv_frozen_continuous instance"""
    return isinstance(dist, rv_frozen) and isinstance(dist.dist, rv_continuous)


def _change_field_representation(
    dataclass_instance: dataclass, representations_to_change: dict[str, Any]
) -> str:
    """Just like the default __repr__ but supports reformatting and replacing some values."""
    final = []
    for current_field in dataclass_instance.__dataclass_fields__.values():
        if not current_field.repr:
            continue
        name = current_field.name
        value = representations_to_change.get(
            name, dataclass_instance.__getattribute__(name)
        )
        final.append(f"{name}={value}")
    return f"{dataclass_instance.__class__.__name__}({', '.join(final)})"


def _create_distribution_representation(distribution: rv_frozen) -> str:
    """Create a readable representation of rv_frozen instances"""
    args = ", ".join([str(a) for a in distribution.args])
    kwargs = ", ".join([f"{k}={v}" for k, v in distribution.kwds])
    params = [a for a in [args, kwargs] if a]
    return f"{distribution.dist.name}({', '.join(params)})"


@dataclass
class ContinuousVariable:
    """A variable with continuous distribution"""

    distribution: Optional[rv_frozen] = None
    lower_bound: Optional[float] = None
    upper_bound: Optional[float] = None
    infinite_bound_probability_tolerance: float = 1e-6

    def __post_init__(self) -> None:
        if self.distribution is None and None in [self.lower_bound, self.upper_bound]:
            raise ValueError(
                "Either the distribution or both "
                "lower_bound and upper_bound have to be set."
            )
        if self.distribution is None:
            self.distribution = uniform(
                self.lower_bound, self.upper_bound - self.lower_bound
            )
        if (
            None not in [self.lower_bound, self.upper_bound]
            and self.lower_bound >= self.upper_bound
        ):
            raise ValueError("lower_bound has to be smaller than upper_bound")
        if not _is_frozen_continuous(self.distribution):
            raise ValueError("Only frozen continuous distributions are supported.")

    def value_of(
        self, probability: Union[float, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """Given a probability or an array of probabilities return the corresponding value(s) using the inverse cdf."""
        values = self.distribution.ppf(probability)
        if self.upper_bound is not None or self.lower_bound is not None:
            return np.clip(values, self.lower_bound, self.upper_bound)
        return values

    def cdf_of(self, value: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Given a value or an array of values return the probability using the cdf."""
        return self.distribution.cdf(value)

    @property
    def finite_lower_bound(self) -> float:
        """Provide a finite lower bound of the variable even if it was not provided by the user."""
        if self.lower_bound is not None:
            return self.lower_bound
        value = self.value_of(0.0)
        if np.isfinite(value):
            return value
        return self.value_of(self.infinite_bound_probability_tolerance)

    @property
    def finite_upper_bound(self) -> float:
        """Provide a finite upper bound of the variable even if it was not provided by the user."""
        if self.upper_bound is not None:
            return self.upper_bound
        value = self.value_of(1.0)
        if np.isfinite(value):
            return value
        return self.value_of(1 - self.infinite_bound_probability_tolerance)

    def __repr__(self) -> str:
        distribution_representation = _create_distribution_representation(
            self.distribution
        )
        return _change_field_representation(
            self, {"distribution": distribution_representation}
        )


@dataclass
class DiscreteVariable:
    """A variable with discrete distribution"""

    distribution: rv_frozen
    value_mapper: Callable[[float], Union[float, int]] = lambda x: x
    inverse_value_mapper: Callable[[float, int], Union[float]] = lambda x: x
    infinite_bound_probability_tolerance: float = 1e-6

    def __post_init__(self) -> None:
        if not _is_frozen_discrete(self.distribution):
            raise ValueError("Only frozen discrete distributions are supported.")
        self.value_mapper = np.vectorize(self.value_mapper)
        self.inverse_value_mapper = np.vectorize(self.inverse_value_mapper)

    def value_of(
        self, probability: Union[float, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """Given a probability or an array of probabilities return the corresponding value(s) using the inverse cdf."""
        values = self.distribution.ppf(probability)
        return self.value_mapper(values)

    def cdf_of(self, values: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Given a value or an array of values return the probability using the cdf."""
        return self.distribution.cdf(self.inverse_value_mapper(values))

    @property
    def finite_lower_bound(self) -> float:
        """Provide a finite lower bound of the variable even if it was not provided by the user."""
        support = self.distribution.support()
        if np.isfinite(support[0]):
            return self.value_mapper(support[0])
        return self.value_of(self.infinite_bound_probability_tolerance)

    @property
    def finite_upper_bound(self) -> float:
        """Provide a finite upper bound of the variable even if it was not provided by the user."""
        support = self.distribution.support()
        if np.isfinite(support[-1]):
            return self.value_mapper(support[1])
        return self.value_of(1 - self.infinite_bound_probability_tolerance)

    def __repr__(self) -> str:
        distribution_representation = _create_distribution_representation(
            self.distribution
        )
        return _change_field_representation(
            self, {"distribution": distribution_representation}
        )


def create_discrete_uniform_variables(
    discrete_sets: list[list[Union[int, float, str]]]
) -> list[DiscreteVariable]:
    """Given sets of possible values, create corresponding discrete variables with equal probability of each value."""
    variables = []
    for discrete_set in discrete_sets:
        n_values = len(discrete_set)
        if n_values < 2:
            raise ValueError("At least two values are required for discrete variables")
        # In the following, it is OK and even advantageous to have a mutable
        # default argument as a very rare occasion. Therefore, we disable inspection.
        # noinspection PyDefaultArgument
        variables.append(
            DiscreteVariable(
                distribution=randint(0, n_values),
                # Don't forget to bind the discrete_set below either by
                # defining a kwarg as done here, or by generating in another
                # scope, e.g. function. Otherwise, the last value of discrete_set
                # i.e. the last entry of discrete_sets will be used for all converters
                # Check https://stackoverflow.com/questions/19837486/lambda-in-a-loop
                # for a description as this is expected python behaviour.
                value_mapper=lambda x, values=sorted(discrete_set): values[int(x)],
                inverse_value_mapper=lambda x, values=sorted(
                    discrete_set
                ): values.index(x),
            )
        )
    return variables


def create_continuous_uniform_variables(
    continuous_lower_bounds: Sequence[float], continuous_upper_bounds: Sequence[float]
) -> list[ContinuousVariable]:
    """Given lower and upper bounds, create uniform variables."""
    if len(continuous_lower_bounds) != len(continuous_upper_bounds):
        raise ValueError(
            "Number of lower bounds has to be equal to the number of upper bounds"
        )
    variables = []
    for lower, upper in zip(continuous_lower_bounds, continuous_upper_bounds):
        variables.append(ContinuousVariable(lower_bound=lower, upper_bound=upper))
    return variables


class Variable(Protocol):

    @property
    def distribution(self) -> rv_frozen:
        """Distribution of the variable"""

    def value_of(
        self, probability: Union[float, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """Given a probability or an array of probabilities return the corresponding value(s) using the inverse cdf."""

    def cdf_of(self, value: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        """Given a value or an array of values return the probability using the cdf."""

    @property
    def finite_lower_bound(self) -> float:
        """Provide a finite upper bound of the variable even if it was not provided by the user."""

    @property
    def finite_upper_bound(self) -> float:
        """Provide a finite upper bound of the variable even if it was not provided by the user."""


def create_variables_from_distributions(
    distributions: list[rv_frozen],
) -> list[Variable]:
    """Given a list of distributions, create the corresponding continuous or discrete variables."""
    variables = []
    for dist in distributions:
        if _is_frozen_discrete(dist):
            variables.append(DiscreteVariable(distribution=dist))
        elif _is_frozen_continuous(dist):
            variables.append(ContinuousVariable(distribution=dist))
        else:
            raise ValueError(
                f"Each distribution must be a frozen discrete or continuous type, got {type(dist)}"
            )
    return variables



from functools import partial

import numpy as np
from scipy.stats import uniform

from experiment_design.experiment_designer import ExperimentDesigner
from experiment_design.optimize import random_search
from experiment_design.scorers import Scorer
from experiment_design.variable import DesignSpace, VariableCollection


class RandomSamplingDesigner(ExperimentDesigner):
    """Create or extend a design of experiments (DoE) by randomly sampling from the variable distributions."""

    def _create(
        self,
        variables: DesignSpace,
        sample_size: int,
        scorer: Scorer,
        initial_steps: int,
        final_steps: int,
        verbose: int,
    ) -> np.ndarray:
        steps = initial_steps + final_steps
        return random_search(
            creator=partial(sample_from, variables, sample_size),
            scorer=scorer,
            steps=steps,
            verbose=verbose,
        )

    def _extend(
        self,
        old_sample: np.ndarray,
        variables: DesignSpace,
        sample_size: int,
        scorer: Scorer,
        initial_steps: int,
        final_steps: int,
        verbose: int,
    ) -> np.ndarray:
        steps = initial_steps + final_steps
        return random_search(
            creator=partial(sample_from, variables, sample_size),
            scorer=scorer,
            steps=steps,
            verbose=verbose,
        )


def sample_from(variables: VariableCollection, sample_size: int) -> np.ndarray:
    """
    Sample from the distributions of the variables.

    :param variables: Determines the dimensions of the resulting sample.
    :param sample_size: the number of points to be created.
    :return: Sample matrix with shape (len(variables), samples_size).
    """
    doe = uniform(0, 1).rvs((sample_size, len(variables)))
    if not isinstance(variables, DesignSpace):
        variables = DesignSpace(variables)
    return variables.value_of(doe)



from experiment_design.variable.variable import (
    ContinuousVariable,
    DiscreteVariable,
    Variable,
    create_continuous_uniform_variables,
    create_discrete_uniform_variables,
    create_variables_from_distributions,
)
from experiment_design.variable.space import DesignSpace, VariableCollection
